* Subroutines for Quicksort and Binary search
*
* pjwitte 2o23+
* All code is my own work except as and where stated
*
* V0.01
*
* Compatible with QDOS, Minerva and SMSQ/E
*


        section code

        xdef cmpel

        include dev8_keys_qlv


* QS idx_base, cnt, offs(), types%(), cmps%()
*
* where
*   each index item is a long word pointer to a record in a heap
*   offs() is an array containing offsets to each field within the record
*   types%() is an array holding the types of corresponding fields
*   cmps%() holds each field's sort direction and comparison type
*     strings: +/- (1..4) => Qdos comparison type 0..3, a..Z or Z..a
*     for numbers cmps% is ignored
*
* Types:
*
*   sbyte = $00                  - signed byte
*   ubyte = $02                  - unsigned byte
*   sword = $04                  - signed 16 bit word
*   uword = $06                  - unsigned 16 bit word
*   slong = $08                  - signed 32 bit long word
*   ulong = $0A                  - unsigned 32 bit long word
*   fltpt = $0C                  - 48 bit float
*   qstrg = $0E                  - Q-string (len.w + bytes)
*
* Offsets:
*
* These are offsets from the record pointer, rp, with rp = index + n * 4
* Offsets within the records will depend on the application.
* A negative offset is, when negated, a pointer to a long word pointer to
* some data of the type given in the types% array.
*

*
cmpel
* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
*
* Compare elements by type:
*
*  d0   =               result -1 (X < Y), 0 (X = Y), 1 (X > Y)
*  d5.l = type          saved   direction.comp type | variable type.w
*  a0  -> X             saved
*  a1  -> Y             saved
*
* Certain actions will have been performed before we get here:
*  Variable types will have been checked for validity and range
*  Pointer to pointer: The actual data will be pointed to;
*  C-ctrings: the pair of strings will have been converted to q-strings in
*  a buffer pointed to by a0 and a1.
*

cmp_reg reg a0-a3/d5


        movem.l cmp_reg,-(a7)

        move.w tab(pc,d5.w),d0
        jsr tab(pc,d0.w)            jump to it

return
        movem.l (a7)+,cmp_reg

exit
        rts

*
tab
        dc.w 0              dummy
        dc.w sbyte_cmp-tab
        dc.w ubyte_cmp-tab
        dc.w sword_cmp-tab
        dc.w uword_cmp-tab
        dc.w slong_cmp-tab
        dc.w ulong_cmp-tab
        dc.w fltpt_cmp-tab
        dc.w qstrg_cmp-tab

*
sbyte_cmp
        move.b (a1),d0
        cmp.b (a0),d0
        bra.s signed

*
sword_cmp
        move.w (a1),d0
        cmp.w (a0),d0
        bra.s signed

*
slong_cmp
        move.l (a1),d0
        cmp.l (a0),d0

signed
        beq.s eq
        blt.s mi
        bgt.s pl

*
ubyte_cmp
        move.b (a1),d0
        cmp.b (a0),d0
        bra.s unsigned

*
uword_cmp
        move.w (a1),d0
        cmp.w (a0),d0
        bra.s unsigned

*
ulong_cmp
        move.l (a1),d0
        cmp.l (a0),d0

unsigned
        beq.s eq
        bcs.s mi
        bhi.s pl

*
qstrg_cmp
         move.w sr,d0                   save current ssr
         trap #0                         freeze for Qdos

         movem.l d5/a6,-(a7)
         suba.l a6,a0                   neutralise a6
         suba.l a6,a1
         swap d5
         move.b d5,d0                   Qdos comparison type
         move.w ut.cstr,a2              jsr ut.cstr*3+qlv.ofs+2
         jsr (a2)

         movem.l (a7)+,d5/a6
         move.w d0,sr                   unfreeze

*
dora
        tst.l   d5                      up or down?
        bmi.s dora_chs

        tst.l d0
        rts

dora_chs
        neg.l d0
        rts

*
fltpt_cmp                               ; cmp.f f1,f0; where
        move.l 2(a0),d0
        cmp.l 2(a1),d0
        bvc.s flt_ss                    sign same

        tst.l d0                        signs different so
        bra.s flt_done                   return comparison

flt_ss
        tst.l d0                        signs positive?
        bmi.s flt_neg

        move.w 0(a0),d1
        cmp.w 0(a1),d1                  how do exps compare?
        bne.s flt_done                   different magnitude

        cmp.l 2(a1),d0                   same magnitude

flt_done
        bmi.s mi
        beq.s eq

pl
        moveq #1,d0
        bra.s dora

eq
        moveq #0,d0
        bra.s dora

mi
        moveq #-1,d0
        bra.s dora

flt_neg
        move.w 0(a1),d1
        cmp.w 0(a0),d1
        bne.s flt_done

        cmp.l 2(a1),d0
        bra.s flt_done

*
        end
